/*
 * CJsonComLayer.cpp
 *
   Version History:
   2016-11-07 Ernst Murnleitner, Awite Bioenergie GmbH: Initial Version

 */

#include "jsoncomlayer.h"
#ifdef FORTE_ENABLE_GENERATED_SOURCE_CPP
#include "jsoncomlayer_gen.cpp"
#endif
#include <forte_any.h>
#include <forte_string.h>


// the following 3 lines of includes beginning with json_spirit refer to source code taken from json_spirit 4.08
// json_spirit: ((c) John W. Wilkinson 2007 - 2014, distributed under the MIT license)
// Distributed under the MIT License:
#include "json_spirit_value_implementation.h"
#include "json_spirit_reader_implementation.h"
#include "json_spirit_writer_implementation.h"


/// get numeric value from value
template <class A> A json_get_numeric(const json_spirit::mValue& value, A dDefault, bool& bDefaultWasSet)
{

    bDefaultWasSet = false;
    try
    {
      A d = dDefault;
      switch (value.type())
      {
      case json_spirit::str_type:
      {
        std::stringstream ss;
        ss.precision(std::numeric_limits<double>::digits10 + 2);
        ss << value.get_str();
        ss >> d;
      }
      break;

      case json_spirit::bool_type:
      {
        d = value.get_bool();
      }
      break;

      case json_spirit::int_type:
      {
        d =  A(value.get_int());
      }
      break;

      case json_spirit::real_type:
      {
        d =  A(value.get_real());
      }
      break;

      default:
      {
        d = dDefault;
        bDefaultWasSet = true;
      }
      break;
      } // end of switch (value.type())


      return d;
    }
    catch (std::runtime_error&)
    {
      std::cerr << "runtime_error: json_get";
      bDefaultWasSet = true;
      return dDefault;
    }
}


enum EJsonValueFromString {
  json_from_string_strict, // be strict according to json standard
  json_from_string_convert_to_smallest, // convert string content to the smallest possible Value type, e.g. "1.0" becomes bool;
  json_from_string_strict_then_smallest, // convet according to the json standard and try to convert to smallest if the first is not possible
};



namespace forte {
namespace com_infra {



std::string getLeftOf( const std::string & source, const char * s, bool bMostRight = false )
{
  std::string sRes;
  std::string str = s;
  size_t iPos = std::string::npos;
  if ( bMostRight )
    iPos = source.rfind( str );
  else
    iPos = source.find( str );
  if ( iPos != std::string::npos )
  {
    if ( iPos == 0 )
      return sRes;

    sRes = source.substr( 0, iPos );
  }
  else
    return source;

  return sRes;
}


std::string getRightOf( const std::string & source, const char * s, bool bMostRight = false )
{
  if ( ( s == 0 ) || ( *s == '\0' ) )
    return source;  // mur 85



  std::string sRes;
  if (source != "")
  {
    std::string str = s;
    size_t iPos = std::string::npos;
    if (bMostRight)
      iPos = source.rfind( str );
    else
      iPos = source.find( str );

    if ( iPos != std::string::npos )
    {
      if ( ( iPos + str.size() ) >= source.size() )
        return sRes;

      sRes = source.substr( iPos + str.size(), source.size() - iPos - str.size() );
    }
  }
  return sRes;

}

std::string mirror(const std::string& str)
{
  if (!str.size())
  {
    std::string sEmpty;
    return sEmpty;
  }

  std::string s = str;
  int j = 0;
  for (size_t i = str.size(); i > 0; i--, j++)
  {
    s[i-1] = str[j];
  }

  return s;
}


std::string trimm( const std::string & source )
{
  std::string sRes;
  if (source != "")
  {
    size_t iPosLeft = source.find_first_not_of( "\t \r\n" );

    if ( iPosLeft != std::string::npos )
    {
      sRes = source.substr( iPosLeft, source.size() - iPosLeft );


      // und nun rechts:
      sRes = mirror( sRes );
      iPosLeft = sRes.find_first_not_of( "\t \r\n" );
      if ( iPosLeft != std::string::npos )
        sRes = sRes.substr( iPosLeft, sRes.size() - iPosLeft );
      sRes = mirror( sRes );


    }
    else
    {
      // besteht nur aus Leerzeichen oder ist leer

      sRes = "";

    }

    // "" entfernen:
    if ( sRes.size() >= 2)
      if ( sRes[ 0 ] == '"' && sRes[ sRes.size() - 1 ] == '"' )
        sRes = sRes.substr( 1, sRes.size() - 2 );
  }
  return sRes;

}



json_spirit::mValue json_findValue(const json_spirit::mValue& data, const std::string& variableName,
    const json_spirit::mValue& defaultValue)
{
  try
  {

    // json_spirit::mValue v(data);
    json_spirit::mObject obj = data.get_obj();

    if (obj.count(variableName))
    {
      json_spirit::mValue v2 = obj.at(variableName);
      return v2;
    }
  }
  catch (std::runtime_error& err)
  {

    return (defaultValue);
  }

  return (defaultValue);
}



// iec -> json, needed for publishing
bool json_get_from_CIEC_ANY( json_spirit::Value& value, const CIEC_ANY &paDataPoint, EJsonDataType eJsonForcedDataType)
{
  bool bSuccess = true;

  EJsonDataType eBest = json_null;
  double d = 0;
  bool bHandled = true;


  switch(paDataPoint.getDataTypeID())
  {
  case CIEC_ANY::e_BOOL:
    d = ((CIEC_BOOL &) paDataPoint).operator bool();
    eBest = json_bool;
    break;
  case CIEC_ANY::e_BYTE:
  case CIEC_ANY::e_USINT:
  case CIEC_ANY::e_UINT:
  case CIEC_ANY::e_WORD:
  case CIEC_ANY::e_UDINT:
  case CIEC_ANY::e_DWORD:
  case CIEC_ANY::e_DINT:
  case CIEC_ANY::e_SINT:
  case CIEC_ANY::e_INT:
    d = ((CIEC_DINT &) paDataPoint).operator TForteInt32();
    eBest = json_int;
    break;

  case CIEC_ANY::e_REAL:
    d = (double)((CIEC_REAL &) paDataPoint).operator TForteFloat();
    eBest = json_real;
    break;
  case CIEC_ANY::e_LREAL:
    d = (double)((CIEC_LREAL &) paDataPoint).operator TForteDFloat();
    eBest = json_real;
    break;
  default:
    bHandled = false;
    eBest = json_null;
    break;

  } // end switch


  if (bHandled)
  {
    switch(eJsonForcedDataType)
    {
    case json_any:
    case json_auto:
    case json_number:
    {
      switch (eBest)
      {
      case json_bool:
        value = json_spirit::Value(bool(d));
        break;
      case json_int:
#ifdef FORTE_USE_64BIT_DATATYPES
        value = json_spirit::Value(int64_t(d));
#else
        value = json_spirit::Value(int32_t(d));
#endif
        break;
      case json_real:
        value = json_spirit::Value(d);
        break;
      default:
        break;
      } // end switch eBest
    }
    break;
    case json_bool:
    {
      if (d == 0.0 || d == 1.0)
      {
        value = json_spirit::Value(bool(d));
      }
      else
      {
        bSuccess = false;
      }
    }
    break;
    case json_int:
    {
#ifdef FORTE_USE_64BIT_DATATYPES
      if (d == int64_t(d))

      {
        value = json_spirit::Value(int64_t(d));
      }
#else
      if (d == int32_t(d))

      {
        value = json_spirit::Value(int32_t(d));
      }
#endif
      else
      {
        bSuccess = false;
      }
    }
    break;
    case json_real:
    {
      value = json_spirit::Value(d);

    }
    break;
    default:
      bSuccess = false;
      break;
    } // end switch


    if (! bSuccess && eJsonForcedDataType == json_any)
    { // don't mind if accuracy is lost
      if (d == 0 || d == 1)
      {
        value = json_spirit::Value(bool(d));
      }
#ifdef FORTE_USE_64BIT_DATATYPES
      else if (d == int64_t(d))

      {
        value = json_spirit::Value(int64_t(d));
      }
#else
      else if (d == int32_t(d))

      {
        value = json_spirit::Value(int32_t(d));
      }
#endif
      else
      {
        value = json_spirit::Value(d);

      }
      bSuccess = true;
    }

  } // end if bHandled


  if (! bHandled)
  { // string?
    switch(eJsonForcedDataType)
    {
    case json_auto:
    case json_string:
    case json_any:
    {

      switch(paDataPoint.getDataTypeID())
      {
      case CIEC_ANY::e_STRING:
      case CIEC_ANY::e_WSTRING:
        value = json_spirit::Value( ((CIEC_ANY_STRING &) paDataPoint).getValue());
        break;
      default:
      {
        switch (eBest)
        {
        case json_bool:
        case json_int:
        case json_real:
        {
          std::stringstream ss;
          ss << d;
          value = json_spirit::Value(ss.str());
          bHandled = true;
          bSuccess = true;
        }
        break;
        default:
        {
          char buf[1024];
          paDataPoint.toString(buf,1023);
          buf[1023] = 0;
          value = json_spirit::Value(buf);
          bHandled = true;
          bSuccess = true;
        }
        break;
        } // end switch eBest
      }
      break;
      } // end switch (paDataPoint.getDataTypeID())
    }
    break;
    default:
      break;
    } // end switch (eJsonForcedDataType)

  } // end if ! bHandled

  if (! bHandled)
  {
    bSuccess = false;
  }
  return bSuccess;
}



//  json -> iec, needed for subscribing, receive data
bool json_set_CIEC_ANY(const json_spirit::mValue& value, CIEC_ANY &paDataPoint, EJsonDataType eJsonForcedDataType)
{

  bool retVal = true;
  switch (paDataPoint.getDataTypeID()){
  case CIEC_ANY::e_BOOL:
    if(value.type() == json_spirit::bool_type){
      ((CIEC_BOOL &) paDataPoint) =  value.get_bool();
    }else{
      retVal = false;
    }
    break;
  case CIEC_ANY::e_SINT:
  case CIEC_ANY::e_INT:
  case CIEC_ANY::e_DINT:
  case CIEC_ANY::e_USINT:
  case CIEC_ANY::e_UINT:
  case CIEC_ANY::e_BYTE:
  case CIEC_ANY::e_WORD:
    //Everything smaller or equal 32Bit will be handled here
    if(value.type() == json_spirit::int_type ){
      ((CIEC_DINT &) paDataPoint) = value.get_int();
    }else{
      retVal = false;
    }
    break;
  case CIEC_ANY::e_LINT:
  case CIEC_ANY::e_UDINT:
  case CIEC_ANY::e_ULINT:
  case CIEC_ANY::e_DWORD:
  case CIEC_ANY::e_LWORD:
    //Everything needing more than 32Bit signed integer will be handled here
    if(value.type() == json_spirit::int_type){
      ((CIEC_LINT &) paDataPoint) = value.get_int();
    }else{
      retVal = false;
    }
    break;
  case CIEC_ANY::e_REAL:
    if(value.type() == json_spirit::real_type){
      ((CIEC_REAL &) paDataPoint) = (TForteFloat) (value.get_real());
    }else{
      retVal = false;
    }
    break;
  case CIEC_ANY::e_LREAL:
    if(value.type() == json_spirit::real_type){
      ((CIEC_LREAL &) paDataPoint) = value.get_real();
    }else{
      retVal = false;
    }
    break;
  case CIEC_ANY::e_STRING:
  case CIEC_ANY::e_WSTRING:
    if(value.type() == json_spirit::str_type){
      ((CIEC_ANY_STRING &) paDataPoint) = value.get_str().c_str();
    }else{
      retVal = false;
    }
    break;
  default:
    //TODO think on how to handle these e_DATE, e_TIME_OF_DAY, e_DATE_AND_TIME, e_TIME
    retVal = false;
    break;
  }

  if (retVal == false && eJsonForcedDataType == json_any)
  { // try again

    retVal = true;
    bool bDefaultWasSet = false;
    double d = json_get_numeric(value,0,bDefaultWasSet);
    if (! bDefaultWasSet)
    {
      // we got a numeric value
      switch (paDataPoint.getDataTypeID()){
      case CIEC_ANY::e_BOOL:

        ((CIEC_BOOL &) paDataPoint) =  d;

        break;
      case CIEC_ANY::e_SINT:
      case CIEC_ANY::e_INT:
      case CIEC_ANY::e_DINT:
      case CIEC_ANY::e_USINT:
      case CIEC_ANY::e_UINT:
      case CIEC_ANY::e_BYTE:
      case CIEC_ANY::e_WORD:
        //Everything smaller or equal 32Bit will be handled here

        ((CIEC_DINT &) paDataPoint) = (TForteInt32)d;

        break;
      case CIEC_ANY::e_LINT:
      case CIEC_ANY::e_UDINT:
      case CIEC_ANY::e_ULINT:
      case CIEC_ANY::e_DWORD:
      case CIEC_ANY::e_LWORD:
        //Everything needing more than 32Bit signed integer will be handled here

        ((CIEC_LINT &) paDataPoint) = (TForteInt64)d;

        break;
      case CIEC_ANY::e_REAL:

        ((CIEC_REAL &) paDataPoint) = (TForteFloat)d;

        break;
      case CIEC_ANY::e_LREAL:

        ((CIEC_LREAL &) paDataPoint) = d;

        break;
      case CIEC_ANY::e_STRING:
      case CIEC_ANY::e_WSTRING:
      {
        ((CIEC_ANY_STRING &) paDataPoint) = value.get_str().c_str();

        break;
      }
      default:
        //TODO think on how to handle these e_DATE, e_TIME_OF_DAY, e_DATE_AND_TIME, e_TIME
        retVal = false;
        break;
      }
    }

  }

  return retVal;
}




CJsonComLayer::CJsonComLayer(CComLayer* pa_poUpperLayer, CCommFB * pa_poFB) : CComLayer(pa_poUpperLayer, pa_poFB) {
  // TODO Auto-generated constructor stub

}

CJsonComLayer::~CJsonComLayer() {
  // TODO Auto-generated destructor stub
}
void CJsonComLayer::closeConnection(){
  if(0 != m_poBottomLayer){
    m_poBottomLayer->closeConnection();
  }
}

EComResponse CJsonComLayer::sendData( void *pa_pvData, unsigned int pa_unSize){

  size_t vecnum = m_jsonNameVector.size();
  unsigned int datanum =  m_poFb->getNumSD();
  if (pa_unSize < datanum)
  {
    datanum = pa_unSize;
  }

  TConstIEC_ANYPtr apoSDs = static_cast<TConstIEC_ANYPtr>(pa_pvData);

  std::map<std::string, json_spirit::Object * > m;
  std::vector<std::map<size_t, std::string > > indexMapVec; // vector: i, map: levels

  m[""] = new (json_spirit::Object);

  // get the values first and add immediately if no hierarchy
  size_t maxLevelsOfAll = 0;
  std::vector< json_spirit::Value > valuevec;
  for (size_t i = 0; i <  datanum; i++)
  { // for each var
    if (i >= vecnum)
    {
      break;
    }
    TConstIEC_ANYPtr apo = &(apoSDs[i]);
    json_spirit::Value v;
    /*bool bSuccess = */ json_get_from_CIEC_ANY(v, *apo, m_jsonTypeVector[i]);
    valuevec.push_back( v);
    size_t levels = m_jsonHierarchyVectorVector[i].size();
    if (levels > maxLevelsOfAll)
    {
      maxLevelsOfAll = m_jsonHierarchyVectorVector[i].size();
    }
    if (levels == 0)
    {
      // we have no subobjects, add value direct to the root object
      json_spirit::Object * p = m[""];
      if (p)
      {
        p->push_back(json_spirit::Pair(m_jsonNameVector[i],valuevec[i]));
      }
    }
  }

  // build matrix with json objects

  for (size_t i = 0; i <  datanum; i++)
  { // for each var
    if (i >= vecnum)
    {
      break;
    }
    std::map<size_t, std::string > indexMap;
    indexMapVec.push_back(indexMap);
  }
  for (size_t i = 0; i <  datanum; i++)
  { // for each var
    if (i >= vecnum)
    {
      break;
    }

    const std::vector<std::string > & levelvec = m_jsonHierarchyVectorVector[i];
    std::string s;

    for (size_t level = 0; level < levelvec.size(); level++)
    {

      //json_spirit::Object * parent = m[s];
      std::string s_parent = s;

      s += "/";
      s += levelvec[level];
      bool bExisting = m.count(s);
      if (bExisting)
      {
        // stop next lower level
        //(indexMapVec[i])[level] = "-";

      }
      else
      {
        json_spirit::Object * p = new (json_spirit::Object);
        m[s] = p;

      }
      (indexMapVec[i])[level] = s;

    }
  }




  // fill matrix from the end

  for (size_t level = maxLevelsOfAll; level > 0; level --)
  {
    // A add value
    for (size_t i = 0; i <  datanum; i++)
    { // for each var
      if (i < vecnum)
      {
        const std::vector<std::string > & levelvec = m_jsonHierarchyVectorVector[i];
        if (level <= levelvec.size())
        {
          if (level == levelvec.size())
          {
            std::string s = (indexMapVec[i])[level-1];
            json_spirit::Object * p = m[s];
            // we are at the end, add value
            if (p)
            {
              p->push_back(json_spirit::Pair(m_jsonNameVector[i],valuevec[i]));
            }
          }
        } // end if level <= levelvec.size()
      } // end if i < vecnum
    } // end for i

    // B add to parent
    for (size_t i = 0; i <  datanum; i++)
    { // for each var
      if (i < vecnum)
      {
        const std::vector<std::string > & levelvec = m_jsonHierarchyVectorVector[i];
        if (level <= levelvec.size())
        {
          std::string s = (indexMapVec[i])[level-1];
          json_spirit::Object * p = m[s];

          // add the new object to the next lower level
          if (level >= 2)
          {
            std::string sparent = (indexMapVec[i])[level-2];
            json_spirit::Object * parent = m[sparent];
            if (parent)
            {

              std::string s2 = (indexMapVec[i])[level-1];
              s2 = getRightOf(s2,"/",true);
              parent->push_back(json_spirit::Pair(s2,*p));
            }
          }

          // add the first level to the root level if not already added
          if (level == 1)
          {
            bool bAlreadyAdded = false;
            for (size_t j = 0; j < i; j++)
            {
              if ((indexMapVec[j])[level-1] == (indexMapVec[i])[level-1])
              {
                bAlreadyAdded = true;
                break;
              }
            }

            if (! bAlreadyAdded)
            {
              // add to root object
              json_spirit::Object * root = m[""];
              if (root)
              {
                std::string s = (indexMapVec[i])[level-1];
                s = getRightOf(s,"/",true);
                root->push_back(json_spirit::Pair(s,*p));
              }
            }
          } // end if level == 1
        } // end if level <= levelvec.size()
      } // end if i < vecnum
    } // end for i
  } // end for level

  std::string strPayload = json_spirit::write(*(m[""]), json_spirit::raw_utf8);
  m_poBottomLayer->sendData(const_cast<void *>(reinterpret_cast<const void*>(strPayload.c_str())), (unsigned int) strPayload.size());

  std::map<std::string, json_spirit::Object * >::iterator it = m.begin();
  while (it != m.end())
  {
    if (it->second)
    {
      delete it->second;
    }
    ++ it;
  }
  return e_ProcessDataOk;
}

EComResponse CJsonComLayer::recvData( const void *pa_pvData, unsigned int pa_unSize){

  EComResponse result = e_ProcessDataOk;
  if (0 == m_poTopLayer)
  {
    const size_t vecnum = m_jsonNameVector.size();
    const unsigned int datanum =  m_poFb->getNumRD();

    char buf[pa_unSize + 1];
    memcpy(buf,pa_pvData,pa_unSize);
    buf[pa_unSize] = 0;

    std::string strPayload = (const char *) buf;
    json_spirit::mValue payl;
    json_spirit::read(strPayload, payl);


    for (size_t i = 0; i <  vecnum; i++)
    {
      if (i < datanum)
      {
        try
        {
          TIEC_ANYPtr apo = (m_poFb->getRDs()+i);
          json_spirit::mValue value;
          if (m_jsonHierarchyVectorVector[i].size())
          { // the value is in subobjects
            // first level:
            value = json_findValue(payl,(m_jsonHierarchyVectorVector[i])[0] , json_spirit::mValue());

            // next levels:
            for (size_t level = 1; level < m_jsonHierarchyVectorVector[i].size(); level++)
            {

              std::string slevel = (m_jsonHierarchyVectorVector[i])[level];
              value = json_findValue(value,slevel, json_spirit::mValue());

            }

            { // last level, get the value
              value = json_findValue(value,m_jsonNameVector[i], json_spirit::mValue());
            }
          }
          else
          { // search in base level only
            value = json_findValue(payl, m_jsonNameVector[i], json_spirit::mValue());
          }

          bool bSetDatapoint = json_set_CIEC_ANY(value, *apo, m_jsonTypeVector[i]);
          if (! bSetDatapoint)
          {
            result = e_ProcessDataDataTypeError;
          }
        }
        catch (std::runtime_error&)
        {
          std::cerr << "runtime_error: json recv data";
          result = e_ProcessDataDataTypeError;
        }
      }
      else
      {
        break;
      }
    }
  }

  return result;
}

EComResponse CJsonComLayer::openConnection(char * pa_acLayerParameter){
  // read parameterised object names expected to be included in the mqtt payload, e.g. "name1, name2, ..."

  std::string s = pa_acLayerParameter;

  while(s != "")
  {
    EJsonDataType json_vartype = json_auto;
    std::string name = trimm(getLeftOf(s,","));
    if (name.size())
    {
      if (name.find(":") != std::string::npos)
      {
        std::string topic = getLeftOf(name,":");

        if (topic.find_first_not_of("0123456789") == std::string::npos)
        {
          // ok, there is a number > 0
          size_t index1 = atoi(topic.c_str());
          name = getRightOf(name,":");

          if (name.find(":") != std::string::npos)
          {

            std::string r = getRightOf(name,":",true);
            bool bFoundType = true;
            if (r == "string")
            {
              json_vartype = json_string;
            }
            else if (r == "number")
            {
              json_vartype = json_number;
            }
            else if (r == "real")
            {
              json_vartype = json_real;
            }
            else if (r == "int")
            {
              json_vartype = json_int;
            }
            else if (r == "bool")
            {
              json_vartype = json_bool;
            }
            else if (r == "auto")
            {
              json_vartype = json_auto;
            }
            else if (r == "null")
            {
              json_vartype = json_null;
            }
            else if (r == "any")
            {
              json_vartype = json_any;
            }
            else
            {
              bFoundType = false;
            }

            if (bFoundType)
            {
              name = getLeftOf(name,":", true);
            }
          } // end if find second ":"

          std::vector<std::string > hierarchy;
          // extract object hierarchy from name
          while (name.find("/") != std::string::npos)
          {
            std::string left = getLeftOf(name,"/");
            name = getRightOf(name,"/");
            hierarchy.push_back(left);
          } // end while


          if ((m_jsonNameVector.size() + 1) == index1)
          {
            m_jsonHierarchyVectorVector.push_back(hierarchy);
            m_jsonNameVector.push_back(name);
            m_jsonTypeVector.push_back(json_vartype);
          }
          else if (index1 > 0 && index1 < 100)
          {
            while (m_jsonNameVector.size() < index1)
            { // fill up with empty elements
              std::vector<std::string > dummy;
              m_jsonHierarchyVectorVector.push_back(dummy);
              m_jsonNameVector.push_back(std::string(""));
              m_jsonTypeVector.push_back(json_null);
            }
            m_jsonHierarchyVectorVector[index1-1] = hierarchy;
            m_jsonNameVector[index1-1] = name;
            m_jsonTypeVector[index1-1] = json_vartype;
          } // end index
        } // end if topic contains only numbers
      } // end if find first ":"
    } // end if name.size()
    s = getRightOf(s,",");
  }  // while s
  // no object names in parameter list
  const size_t iNumParams = m_jsonNameVector.size();
  if (iNumParams == 0)
  {
    return e_InitTerminated;
  }
  // more object names than input/or output is not allowed
  switch (m_poFb->getComServiceType()){
  case e_Client:
  case e_Publisher:
    if(iNumParams > m_poFb->getNumSD()) {
      return e_InitTerminated;
    }
    break;
  case e_Server:
  case e_Subscriber:
    if(iNumParams > m_poFb->getNumRD()) {
      return e_InitTerminated;
    }
    break;
  }

  return e_InitOk;
}


EComResponse CJsonComLayer::processInterrupt() {
  return e_ProcessDataOk;
}


} /* namespace com_infra */
} /* namespace forte */
